package ajs.tools;

import ij.*;
import ij.gui.*;
import ij.measure.ResultsTable;
import ij.plugin.Duplicator;
import ij.plugin.PlugIn;
import ij.plugin.ZProjector;
import ij.process.ImageStatistics;
import ij.text.TextPanel;

public class Normalize_Brightness implements PlugIn {

	@Override
	public void run(String arg) {
		String defaultNormalizationChs="Normalize to this channel";
		String defaultNormalizationZ="Change all based on Z-proj";

		ImagePlus imp=WindowManager.getCurrentImage();
		int sls=imp.getNSlices(), frms=imp.getNFrames(), chs=imp.getNChannels();
		boolean dosls=(frms==1 && sls>1);

		boolean aballzproj=false, abeachsls=false;
		int ch=imp.getC(), frm=imp.getT(), sl=imp.getZ();
		int stsl=sl-1,endsl=sl;
		int stc=ch-1,endc=ch, stcn=stc, endcn=endc; //n's for normalizing
		if(dosls){stsl=frm-1;endsl=frm;}
		String[] zchoices=new String[] {"Average Intensity","Max Intensity", "Min Intensity","Sum Slices","Standard Deviation","Median"};
		String zchoice="Max Intensity";
		boolean useResults=false;
		int nResults=0;
		TextPanel results=null;
		if(WindowManager.getWindow("Results")!=null) {
			results=IJ.getTextPanel();
			nResults=results.getLineCount();
		}
		int[] whichsls=new int[] {1,1};
		String abchopts=defaultNormalizationChs;
		if(chs>1 || (sls>1 &&  frms>1)){
			GenericDialog gd=new GenericDialog("Auto Brightness");
			gd.addMessage("Ch: "+(ch)+" Sl: "+sl+" Fr: "+frm+" on stack: "+imp.getTitle());
			if(chs>1) {
				gd.addNumericField("Ch:", ch, 0);
				gd.addChoice("Multiple ch options:",new String[] {"Do only one channel","Do all channels separately","Normalize to this channel"},defaultNormalizationChs);
			}
			if(nResults==frms || (dosls && nResults==sls))gd.addCheckbox("Use Means from Results?",false);
			if(sls>1 && frms>1) {
				gd.addChoice("Adjust brightness over slices or frames?",new String[] {"Frames","Slices"},"Frames");
				gd.addChoice("Do whole stack?",new String[] {"Only current fr/sl","Repeat individually for all frs/sls","Change all based on Z-proj"},defaultNormalizationZ);
				zchoice=zchoices[(int)ij.Prefs.get("zproject.method", 1)];
				gd.addChoice("If z-proj, how?",zchoices,zchoice);
				gd.addStringField("If z-proj, which sls:?", "1-"+sls);
			}
			gd.showDialog();
			if(gd.wasCanceled())return;
			if(chs>1) {
				ch=(int)gd.getNextNumber();
				imp.setPosition(ch, sl, frm);
				stc=ch-1; endc=ch; stcn=stc; endcn=endc; //n's for normalizing
				abchopts=gd.getNextChoice();
				if(abchopts=="Do all channels separately"){stc=0; endc=chs;}
				if(abchopts=="Normalize to this channel"){stcn=0; endcn=chs;}
			}
			if(nResults==frms || (dosls && nResults==sls))useResults=gd.getNextBoolean();
			String aballslschoice="";
			if(sls>1 && frms>1) {
				dosls=(gd.getNextChoice()=="Slices");
				aballslschoice=gd.getNextChoice();
				zchoice=gd.getNextChoice();
				whichsls=AJ_Utils.parseRange(gd.getNextString());
				abeachsls=(aballslschoice=="Repeat individually for all frs/sls");
				aballzproj=(aballslschoice=="Change all based on Z-proj");
				if(dosls){stsl=frm-1; endsl=frm;}
				if(abeachsls){stsl=0;if(dosls){endsl=frms;}else{endsl=sls;}}
			}
		}
		int endfrms=dosls?sls:frms;
		double[] means=new double[endfrms];
		double maxmean=0;
		int alldim=1;if(aballzproj){if(dosls){alldim=frms;}else{alldim=sls;}}
		if(useResults){
			for(int i=0;i<nResults;i++){
				ResultsTable rt=results.getResultsTable();
				means[i]=rt.getValueAsDouble(rt.getColumnIndex("Mean"), i);
				maxmean=Math.max(means[i],maxmean);
			}
		}
		Roi roi=imp.getRoi();
		boolean selec=roi!=null;
		if(selec) {IJ.run("Select None");}
		if(!dosls && aballzproj && !useResults){
			IJ.run("Z Project...", "start="+whichsls[0]+" stop="+whichsls[1]+" projection=["+zchoice+"] all");
			ImagePlus zimp=WindowManager.getCurrentImage();
			if(chs>0)zimp.setPosition(stc+1,imp.getZ(),imp.getT());
			if(selec) IJ.run("Restore Selection");
			for(int j=0;j<endfrms;j++){
				zimp.setPosition(ch,1,j+1);
				ImageStatistics imgstat=ImageStatistics.getStatistics(zimp.getProcessor(), 127, zimp.getCalibration());
				//double area=imgstat.area;
				double mean=imgstat.mean;
				maxmean=Math.max(mean,maxmean);
				means[j]=mean;
				IJ.showStatus("Measuring means...");
				IJ.showProgress((double)(j+1)/(double)endfrms);
			}
			zimp.close();
		}
		double n=0;
		for(int chi=stc;chi<endc;chi++){
			for(int i=stsl;i<endsl;i++) {
				if(dosls || !aballzproj && !useResults){
					double mean;
					if(selec) run("Restore Selection");
					for(int j=0;j<endfrms;j++){
						int sli=i, frmj=j;
						if(dosls) {sli=j; frmj=i;}
						if(aballzproj){
							Duplicator dup=new Duplicator();
							ImagePlus did;
							if(dosls)did=dup.run(imp, chi+1, chi+1, sli+1, sli+1, 1, frms);
							else did=dup.run(imp, chi+1, chi+1, 1, sls, frmj+1, frmj+1);
							IJ.run("Z Project...", "projection=["+zchoice+"]");
							ImagePlus zd=WindowManager.getCurrentImage();
							ImageStatistics imgstat=ImageStatistics.getStatistics(zd.getProcessor(), 127, zd.getCalibration());
							mean=imgstat.mean;
							zd.close();
							did.close();
						} else {
							imp.setPosition(chi+1,sli+1,frmj+1);
							ImageStatistics imgstat=ImageStatistics.getStatistics(imp.getProcessor(), 127, imp.getCalibration());
							mean=imgstat.mean;
						}
						//print("ch:"+chi+" sl:"+sli+" frm:"+frmj);
						maxmean=Math.max(mean,maxmean);
						means[j]=mean;
						IJ.showStatus("Measuring means...");
						IJ.showProgress(n++/(double)(endfrms*(endsl-stsl)*(endc-stc)*(endcn-stcn)));
					}
				}
				WindowManager.setCurrentWindow(imp.getWindow());
				if(selec) IJ.run("Select None");
				double nn=0;
				ImageStack imst=imp.getStack();
				for(int j=0;j<endfrms;j++){
					for(int k=0;k<alldim;k++){
						int newi=i;
						if(aballzproj){newi=k;}
						int sli=newi, frmj=j;
						if(dosls) {sli=j; frmj=newi;}
						for(int chin=stcn;chin<endcn;chin++){
							int chk=chi;
							if(abchopts=="Normalize to this channel") chk=chin;
							imst.getProcessor(1+chk+sli*chs+frmj*chs*sls).multiply(maxmean/means[j]);
							IJ.showStatus("Normalizing stack (ch"+(chk+1)+"/"+(endcn-stcn)+" sl"+ IJ.pad(k+1,2) +"/"+alldim+" fr"+ IJ.pad(j+1,2) +"/"+endfrms+")...");
							IJ.showProgress(nn++/(double)(endfrms*alldim*(endcn-stcn)));
						}
					}
				}
				IJ.showStatus("Normalization complete.");
				IJ.showProgress(1.0);
			}
		}
		if(selec) run("Restore Selection");
		imp.updateAndDraw();
		}

}
