package ajs.tools;

import ij.plugin.*;
import ij.gui.*;
import ij.measure.Calibration;
import ij.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;

import ij.process.*;
import ij.text.*;

/**
 * This is a template for a plugin that does not require one image
 * (be it that it does not require any, or that it lets the user
 * choose more than one image in a dialog).
 */
public class Thresh_Cell_Transfer2 implements PlugIn, MouseListener, KeyListener, MouseWheelListener {


	final static String version="1.2.0";
	final static String HEADING="Label"+"\t"+"ROI"+"\t"+"Cell#"+"\t"+"Area"+"\t"+"Perimeter"+"\t"+"Circularity"+"\t"+"X"+"\t"+"Y"+"\t"+"Mean"+"\t"+"Slice"+"\t"+"Frame"+"\t"+"Thresh"+"\t"+"Time";
	final static double AREATOLERANCE=0.2;
	final static boolean DEBUG=false;
	final static boolean UPDATE_ACCEPTED=false;
	final static int MAX_POINTS=8;
	final static private Color addColor=new Color(128,64,255);
	final static private Color subColor=new Color(255,0,0);
	static public enum DirectRoiTypes{ 
		SUBTRACTIVE("Subtractive"), ADDITIVE("Additive"), ONLY_WITHIN("Only-within");
		private String string;
		private DirectRoiTypes(String string) {
			this.string=string;
		}
		public String getString() {return string;}
		public static String[] getStrings() {
			DirectRoiTypes[] drt = DirectRoiTypes.values();
			String[] strings=new String[drt.length];
			for(int i=0;i<drt.length;i++)strings[i]=drt[i].getString();
			return strings;
		}
	};
	boolean done=false, acceptedROI=false, gocellcomplete=false, spacepress=false;
	boolean addDirectRoi=false, deleteDirectRoi=false;
	boolean drawCellLabel=false;
	boolean autoAccept=false;
	boolean[] buttonpress=new boolean[3];
	boolean[] postFirstAccept=new boolean[MAX_POINTS];
	boolean showEachPointRoi=true;
	int pointIndex=0;
	ArrayList<Point> xys=new ArrayList<Point>();
	ArrayList<Roi> directRois=new ArrayList<Roi>();
	boolean[] pointIsNeg=new boolean[MAX_POINTS];
	int curX=-1, curY=-1;
	int goback=2, labelsl=0, celln=1;
	double[] currentThresh=new double[MAX_POINTS];
	int wheelfactor = 10;
	int MYLUT=ImageProcessor.RED_LUT;
	private ImagePlus simp;
	private Overlay soverlay;
	private ImagePlus timp;
	String title, endtitle, basetitle;
	ArrayList<String> cellLabels=new ArrayList<String>();
	String cellLabel="Unlabeled";
	AutoWandRoi[] curWand;
	final int[] whfacs= new int[] {1,10,30};
	double[] threshMultiplier;
	TCTPanel tctpanel;
	private MyThread updateCurWandThread=null;
	private MyThread sucwt=null;
	private TctTextWindow results;
	private enum TctLines{
		INFO, OUTPUT, WAND, CURRENTPOINT, THRESH, SLICESLEFT;
	}
	private double[] times;

	//Wand w;

	public void run(String arg) {
		TCT(WindowManager.getCurrentImage());
	}

	public void TCT(ImagePlus imp) {
		simp=imp;
		if (simp==null) {IJ.log("noImage"); return;}
		setup();
		soverlay=new Overlay();
		simp.setOverlay(soverlay);

		tctpanel=new TCTPanel();
		tctpanel.addWindowListener(new WindowAdapter(){  
			public void windowClosing(WindowEvent e) {  
				done=true;
				tctpanel.dispose();  
			}  
		});  
		tctpanel.setVisible(true);

		int sl=simp.getSlice(), fr=simp.getFrame();
		int frms=simp.getNFrames();
		int prevsl=sl, prevfr=fr;
		int prevcelln=-1;
		curX=-1; curY=-1;
		currentThresh[0]=250;
		int xprev=curX, yprev=curY;
		//boolean justfirst;
		threshMultiplier=new double[frms];
		String slsleft="";
		results=new TctTextWindow(basetitle);
		if("Unlabeled".equals(cellLabel)) askCellLabel();
		boolean editing=false;
		//int lastfr=-1;

		//-------loop-------------

		while(!done){
			simp.setT(fr);
			if(celln!=prevcelln){
				slsleft=""; for(int i=0;i<frms;i++) slsleft+=(i+1)+" ";
				tctpanel.setTextLine(TctLines.SLICESLEFT, "Slices left:"+slsleft);
				xys.clear();
				directRois.clear();
				pointIndex=0;
				tctpanel.resetWP(1);
				curX=-1; curY=-1;
				xprev=curX; yprev=curY;
				//lastfr=-1;
				curWand=new AutoWandRoi[frms];
				for(int i=0;i<MAX_POINTS;i++) {
					postFirstAccept[i]=false;
					if(i>0)currentThresh[i]=currentThresh[0];
				}
				simp.getProcessor().setThreshold(currentThresh[0], (double) 65535, MYLUT);
				prevsl=simp.getSlice(); prevfr=simp.getFrame();
				prevcelln=celln;
				soverlay.clear();
				simp.updateAndDraw();
			}
			acceptedROI=false;
			while(!acceptedROI && !gocellcomplete) {
				do{
					fr=simp.getFrame(); sl=simp.getSlice();
					timp.setPosition(1,labelsl,fr);
					if(sl!=prevsl || fr!=prevfr) {
						//tctpanel.setTextLine(5, "Sl"+sl+" psl"+prevsl+" fr"+fr+" pfr"+prevfr);
						if(curWand[fr-1]!=null && curWand[fr-1].roi!=null) {
							if(curWand[fr-1].sl!=sl) {
								curWand[fr-1].updateWandRoi();
								updateCurWand(false);
							}
							tctpanel.setTextLine(TctLines.WAND, "Selection Info--  Roi changed: "+((curWand[fr-1]==null)?"NULL":((curWand[fr-1].roi.equals(curWand[fr-1].troi))?"no":"YES"))+
									" Accepted: "+curWand[fr-1].accepted+" Probable: "+curWand[fr-1].probable+" Points: "+curWand[fr-1].cxys.size()+
									(curWand[fr-1].pointFail.equals("")?"":" Failed point#: "+curWand[fr-1].pointFail));
							tctpanel.printOutput(curWand[fr-1].output);
							curWand[fr-1].showBoth();
						}
						//justfirst=true;
					}
					prevsl=sl; prevfr=fr;
					IJ.wait(10);
				}while(sl!=prevsl || fr!=prevfr);

				if(addDirectRoi)addDirectRoi();
				if(deleteDirectRoi)deleteDirectRoi();

				if(((curX!=xprev||curY!=yprev) /* || justfirst*/) && curX!=-1){
					xprev=curX;yprev=curY;
					Point c=new Point(curX,curY);
					if(pointIndex>=xys.size()) {xys.add(c);
					}else {
						xys.set(pointIndex, c);
					}

					if(curWand[fr-1]==null)curWand[fr-1]=new AutoWandRoi(fr);
					else {
						curWand[fr-1].updatePoint();
					}
					curWand[fr-1].showBoth();
					if(!postFirstAccept[pointIndex])updateCurWand(false);

					//if(justfirst && lastfr!=-1 && xys[lastfr-1]!=null ){
					//	//test if new center is farther than the area's equivalent radius away from the old center
					//	sel=doWandRoiByPt(fr);
					//}
					//justfirst=false;
					tctpanel.setTextLine(TctLines.CURRENTPOINT, "Point:"+(pointIndex+1)+" x:"+curX+" y:"+curY+" z:"+sl+" fr:"+fr);
				}
				if(WindowManager.getWindow(title)==null || WindowManager.getWindow(endtitle)==null) {IJ.log("Window Closed"); done=true;}
				if(done) {cleanup(); return;}
				if(((spacepress && goback!=2) || buttonpress[2]&&goback!=1))IJ.wait(10);
				//if(DEBUG)IJ.log("\\Update:"+System.nanoTime()/1000000+"  xy:"+xy.x+" "+xy.y+" aa"+autoAccept+" probable"+(curWand[fr-1]==null?"NA":""+curWand[fr-1].probable)+" rc:"+buttonpress[2]+" gb:"+goback);
				if(curX!=-1 && curY!=-1 && autoAccept && curWand[fr-1].probable && ((spacepress && goback!=2) || buttonpress[2]&&goback!=1)){acceptedROI=true;}
				//if((auto||!firsthit) && x!=-1 && y!=-1 && !firsttime)acceptedROI=true; else IJ.wait(200);

			}

			if(WindowManager.getCurrentImage()!=simp){
				GenericDialog gd = new GenericDialog("Cancel");
				gd.addMessage("Cancel?");
				gd.enableYesNoCancel("Yes", "No");
				gd.showDialog();
				if (gd.wasCanceled()){
					cleanup(); return;}
				else if (gd.wasOKed()){
					cleanup();return;}
			}
			tctpanel.setTextLine(TctLines.SLICESLEFT, "Slices left: "+slsleft);

			if(curWand[fr-1]!=null)curWand[fr-1].setRoi(simp.getRoi());
			if(curWand[fr-1]==null){
				IJ.log("No selection");
			}else if(!gocellcomplete){
				//WindowManager.setWindow(timp.getWindow());
				curWand[fr-1].accept();
				tctpanel.printOutput(curWand[fr-1].output);
				postFirstAccept[pointIndex]=true;
			}

			long startpresstime=System.nanoTime();
			long waittime=200;
			while(spacepress && !gocellcomplete) {
				IJ.wait(20);
				long presstime=System.nanoTime()-startpresstime;
				//if(!firsthit)waittime=100;
				//IJ.log("wait: "+presstime);
				if((presstime/1000000)>waittime){break;}
			}

			boolean cellcomplete=true;
			slsleft="";
			int lowsl=1;
			for(int i=frms;i>0;i--) {
				if(curWand[i-1]==null || !curWand[i-1].accepted) {cellcomplete=false; slsleft=(i)+" "+slsleft; lowsl=i;}
				else slsleft=" ̸"+(i)+((slsleft.startsWith(" ")||slsleft.startsWith(" ̸"))?"":" ")+slsleft;
			}
			if((cellcomplete && !editing) || gocellcomplete){
				gocellcomplete=false;
				editing=false;
				//firsthit=true;
				GenericDialog gd = new GenericDialog("Cancel");
				gd.addMessage("All done with cell #"+celln+"?");
				gd.enableYesNoCancel("Yes", "No");
				gd.showDialog();
				spacepress=false; buttonpress[0]=false; buttonpress[1]=false; buttonpress[2]=false;
				if (gd.wasOKed()){
					timp.setColor(Color.white);
					for(int i=0;i<frms;i++) {
						if(curWand[i]!=null && curWand[i].accepted) {
							curWand[i].draw();
						}else{
							results.append(cellLabel+"\t0\t"+celln+"\tNA\tNA\tNA\tNA\tNA\tNA\tNA\t"+(i+1)+"\t"+currentThresh+"\t"+times[i]);
						}
					}
					timp.setProperty("Info", results.getText());
					timp.setPosition(1,labelsl,fr);
					Overlay tov=timp.getOverlay();
					if(tov!=null)tov.clear();
					timp.updateAndRepaintWindow();
					celln++; fr=0;
					simp.deleteRoi();
					timp.deleteRoi();
				}else {
					if(!gd.wasCanceled())editing=true;
				}
			}
			//lastfr=fr;
			if(fr<frms) fr++; else fr=lowsl;
			results.roin++;
		}
		cleanup();
	}

	private void setup() {

		IJ.setForegroundColor(255,255,255);
		title=simp.getTitle();
		basetitle=title.endsWith(".tif")?title.substring(0,title.length()-4):title;
		if(title.endsWith("-AJTCT") || title.endsWith("-AJTCT.tif")){
			endtitle=title;
			int indfromend;
			if(title.endsWith("-AJTCT")) indfromend=6; else indfromend=10;
			title=endtitle.substring(0,endtitle.length()-indfromend);
			timp=simp;
			simp=WindowManager.getImage(title);
			if(simp==null)simp=WindowManager.getImage(title+".tif");
			if(simp!=null){basetitle=title; title=title+".tif";}
			else{IJ.showMessage("Can't find original image: "+title); return;}
		}
		if(timp==null){
			endtitle=basetitle+"-AJTCT";
			timp=WindowManager.getImage(endtitle);
			if(timp==null) {
				timp=WindowManager.getImage(endtitle+".tif");
				if(timp!=null)endtitle=endtitle+".tif";
			}
		}
		times=Time_Extractor.extractTimes(simp);
		for(int i=0;i<times.length;i++)times[i]-=times[0];
		ImageWindow siw=simp.getWindow();

		if(timp==null) {
			//timp=IJ.createImage(endtitle,"8-bit composite", simp.getWidth(),simp.getHeight(), 2, 1, simp.getNFrames());
			timp=IJ.createHyperStack(endtitle, simp.getWidth(), simp.getHeight(), 2, 1, simp.getNFrames(), 8);
			if(timp==null) { IJ.log("Error - timp did not get created");return;}
			timp.setDisplayMode(IJ.COMPOSITE);
			timp.show();
			//timp.getProcessor().invertLut();
			//Rectangle tbounds=timp.getWindow().getBounds();
			Rectangle sbounds=siw.getBounds();
			//double sizefactor=sbounds.getWidth()/tbounds.getWidth();
			timp.getWindow().setLocationAndSize((int) (sbounds.getX()+sbounds.getWidth()+5),(int) sbounds.getY(),(int) sbounds.getWidth(),65535);
		}
		if(timp==simp){IJ.log("Same window exiting"); done=true; return;}
		if(timp==null){IJ.log("No target window"); done=true; return;}

		double cthresh=simp.getProcessor().getMinThreshold();
		if(cthresh>0)currentThresh[0]=cthresh;

		WindowManager.setWindow(timp.getWindow());
		WindowManager.setWindow(siw);
		preStrip();
		ImageCanvas ic = simp.getCanvas();
		ic.disablePopupMenu(true);
		ic.addMouseListener(this);
		ic.addKeyListener(this);
		siw.removeMouseWheelListener(siw);
		siw.addMouseWheelListener(this);
	}

	private void addLabel(String newlabel){
		if(!cellLabels.contains(newlabel)){
			cellLabels.add(newlabel);
			celln=1;
		}
		labelsl=cellLabels.indexOf(newlabel)+1;
		while(labelsl>timp.getNSlices()){
			WindowManager.setCurrentWindow(timp.getWindow());
			timp.setZ(timp.getNSlices());
			//IJ.run("Add Slice", "add=slice");IJ.wait(200);
			int channels = timp.getNChannels();
			int slices = timp.getNSlices();
			int frames = timp.getNFrames();
			ImageStack stack=timp.getStack();
			for (int t=frames; t>=1; t--) {
				int index = (t-1)*channels*slices + (slices)*channels;
				for (int i=0; i<channels; i++) {
					ImageProcessor ip = stack.getProcessor(1).duplicate();
					ip.setColor(0); ip.fill();
					stack.addSlice(null, ip, index);
				}
			}
			slices++;
			timp.setStack(stack, channels, slices, frames);
		}
		timp.setPosition(1,labelsl,1);
		timp.getStack().setSliceLabel(newlabel,timp.getCurrentSlice());
		timp.updateAndDraw();
	}

	private void askCellLabel(){
		GenericDialog gd = new GenericDialog("Add new cell label");
		ArrayList<String> askLabels=new ArrayList<String>(cellLabels);
		if(!askLabels.contains("Dura"))askLabels.add(0, "Dura");
		if(!askLabels.contains("Pia"))askLabels.add(1, "Pia");
		String def=cellLabels.isEmpty()?askLabels.get(0):cellLabels.get(cellLabels.size()-1);
		if(!cellLabels.isEmpty() && def.equals("Dura"))def="Pia";
		gd.addChoice("Labels: ",askLabels.toArray(new String[askLabels.size()]),def);
		gd.addStringField("Or new label: ", "");
		gd.addCheckbox("Draw Label on AJTCT stack?", drawCellLabel);
		gd.showDialog();
		if (gd.wasCanceled()) return;
		cellLabel=gd.getNextChoice();
		String temp=gd.getNextString();
		if(!temp.equals(""))cellLabel=temp;
		addLabel(cellLabel);
		drawCellLabel=gd.getNextBoolean();
	}

	private void cleanup(){
		tctpanel.dispose();
		if(simp!=null){
			soverlay.clear();
			simp.updateAndDraw();
			ImageCanvas ic=simp.getCanvas();
			if(ic!=null){
				ImageWindow siw=simp.getWindow();
				ic.removeKeyListener(this);
				ic.removeMouseListener(this);
				siw.removeMouseWheelListener(this);
				siw.addMouseWheelListener(siw);
				ic.disablePopupMenu(false);
			}
		}
		IJ.log("Thresh Cell Tranfer complete");
	}

	private void preStrip(){
		if(simp==null)return;
		ImageCanvas ic=simp.getCanvas();
		ImageWindow siw=simp.getWindow();
		KeyListener[] kls=ic.getKeyListeners();
		for(int i=0;i<kls.length;i++) {
			if(kls[i].getClass().getName().startsWith("Thresh_Cell")){
				ic.removeKeyListener(kls[i]);
				IJ.log("Removed old keyL"+kls[i]);
			}
		}
		MouseListener[] mls=ic.getMouseListeners();
		for(int i=0;i<mls.length;i++) {
			if(mls[i].getClass().getName().startsWith("Thresh_Cell")){
				ic.removeMouseListener(mls[i]);
				IJ.log("Removed old mL"+mls[i]);
			}
		}
		MouseWheelListener[] mwls=siw.getMouseWheelListeners();
		boolean hasit=false;
		for(int i=0;i<mwls.length;i++) {
			if(mwls[i].getClass().getName().startsWith("Thresh_Cell")){
				siw.removeMouseWheelListener(mwls[i]);
				IJ.log("Removed old mwL"+mwls[i]);
			}
			if(mwls[i]==siw)hasit=true;
		}
		if(!hasit) {
			IJ.log("Replacing ImageWindow MouseWheel Listener");
			siw.addMouseWheelListener(siw);
		}
	}

	class TctTextWindow{
		TextWindow rtw;
		public int roin;

		public TctTextWindow(String basetitle) {
			String[] restitles=new String[] {"ThreshCellTransfer-"+basetitle+".xls", "ThreshCellTransfer-"+basetitle+".txt", "ThreshCellTransfer-"+basetitle+".csv"};
			for(String restitle : restitles) {
				rtw=(TextWindow) WindowManager.getWindow(restitle);
				if(rtw!=null)break;
			}
			if(rtw==null){
				rtw=new TextWindow(restitles[0],HEADING,"",800,400);
				if(timp!=null && timp.getInfoProperty()!=null) {
					String tresults=timp.getInfoProperty();
					if(tresults!=null && !"".contentEquals(tresults)) {
						if(tresults.startsWith("\n"))tresults=tresults.substring(1);
						rtw.append(tresults);
					}
				}
			}else {
				TextPanel panel=rtw.getTextPanel();
				String phs=panel.getColumnHeadings();
				if(!phs.contentEquals(HEADING)){
					TextWindow temp=new TextWindow(restitles[0],HEADING,"",800,400);
					ij.measure.ResultsTable restable=panel.getResultsTable();
					ij.measure.ResultsTable ttable=temp.getTextPanel().getResultsTable();
					String[] phss=phs.split("\t");
					String[] hs=HEADING.split("\t");
					int n;
					for(int i=0;i<phss.length;i++) {
						n=-1;
						for(int j=0;j<hs.length;j++) {
							if(phss[i].contentEquals(hs[j])) {n=j; break;}
						}
						if(n>-1) for(int j=0;j<restable.size();j++)ttable.setValue(n, j, restable.getValueAsDouble(i, j));
					}
					rtw.close();
					rtw=temp;
				}
			}
			TextPanel panel=rtw.getTextPanel();
			//IJ.log(""+panel.getLineCount());
			if(panel.getLineCount()>0){
				int troin=0;
				celln=0;
				String[] oldline;
				for(int i=0;i<panel.getLineCount();i++) {
					oldline=panel.getLine(i).split("\t");
					if(i==0){cellLabels.add(oldline[0]);}
					else{ if(!cellLabels.contains(oldline[0])) {cellLabels.add(oldline[0]);}}
					if(!oldline[1].startsWith("NA"))troin=(int)Float.parseFloat(oldline[1]);
					if(!oldline[2].startsWith("NA"))celln=(int) Float.parseFloat(oldline[2]);
					if(troin>roin)roin=troin;
					//newline=""+troin+"\t"+celln+"\t"+oldline[2]+"\t"+oldline[3]+"\t"+oldline[4]+"\t"+oldline[5]+"\t"+(int)Float.parseFloat(oldline[6])+"\t"+(int)Float.parseFloat(oldline[7]);
					//panel.setLine(i,newline);
				}
				if(cellLabels.size()>0)cellLabel=cellLabels.get(cellLabels.size()-1);
				celln++;roin++;
				labelsl=cellLabels.size();
				IJ.log("Using open window, roin: "+roin+" celln: "+celln+" Label: "+cellLabel);
			}
		}

		public void append(String text) {rtw.append(text);}
		public String getText() {
			String[] temp=rtw.getTextPanel().getText().split("\n");
			String result="";
			if(temp.length>1)result=temp[1];
			for(int i=2; i<temp.length;i++)result+="\n"+temp[i];
			return result;
		}
	}

	class TCTPanel extends Frame{

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		TextArea ta=null;
		public Choice whichPoint;

		public TCTPanel() {
			super("TCT Control Panel version "+version);
			/*
			 addWindowListener(new WindowAdapter(){
				public void windowClosing(WindowEvent we){
					recorder.stop();
					dispose();
				}
			});
			 */
			this.setFocusable(false);
			setLayout(new GridBagLayout());
			GridBagConstraints c = new GridBagConstraints();
			String[] backstr=new String[] {"Back - None", "Back - RightClick", "Back - Space"};

			ActionListener l=new ActionListener() {

				@Override
				public void actionPerformed(ActionEvent e) {
					switch(e.getActionCommand()){
					case "LUT":
						changeLutType();
						break;
					case "goback":
						changeGoBack();
						((Button)e.getSource()).setLabel(backstr[goback]);
						break;
					case "askCellLabel":
						askCellLabel();
						break;
					case "gocellcomplete":
						gocellcomplete=true;
						break;
					case "changeWheelFactor":
						changeWheelFactor(1);
						((Button)e.getSource()).setLabel("Wheel - "+wheelfactor);
						break;
					case "updateCurWand":
						boolean clear=false;
						for(int i=0;i<simp.getNFrames();i++)if(curWand[i]!=null && curWand[i].accepted) {clear=true; break;}
						updateCurWand(clear);
						break;
					case "autoAccept":
						autoAccept=!autoAccept;
						((Button)e.getSource()).setLabel("Auto Accept - "+(autoAccept?"yes":"no"));
						break;
					case "autoacceptall":
						autoAcceptAll();
						break;
					case "addDirectRoi":
						addDirectRoi=true;
						break;
					case "deleteDirectRoi":
						deleteDirectRoi=true;
						break;
					case "done":
						done=true;
						break;
					default:
						break;
					}

				}

			};

			//c.fill=GridBagConstraints.HORIZONTAL;
			c.weightx=1; c.weighty=1;
			c.gridx=0;c.gridy=0;
			c.gridwidth=1; c.gridheight=1;
			Button b=new Button();
			b.setFocusable(false);
			b.setActionCommand("LUT");
			b.setLabel("LUT");
			b.addActionListener(l);
			if(simp.isComposite())b.setEnabled(false);
			add(b,c);
			c.gridx++; //1
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("goback");
			b.setLabel(backstr[goback]);
			b.addActionListener(l);
			add(b,c);
			c.gridx++; //2
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("askCellLabel");
			b.setLabel("New Cell Label");
			b.addActionListener(l);
			add(b,c);
			c.gridx++; //3
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("gocellcomplete");
			b.setLabel("Cell complete");
			b.addActionListener(l);
			add(b,c);
			c.gridx++; //4
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("changeWheelFactor");
			b.setLabel("Wheel - "+wheelfactor);
			b.addActionListener(l);
			add(b,c);
			c.gridy++; //1
			c.gridx=0;
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("updateCurWand");
			b.setLabel("Update curWand");
			b.addActionListener(l);
			add(b,c);
			c.gridx++; //1
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("autoAccept");
			b.setLabel("Auto Accept - "+(autoAccept?"yes":"no"));
			b.addActionListener(l);
			add(b,c);
			c.gridx++; //2
			whichPoint=new Choice();
			whichPoint.setFocusable(false);
			resetWP(1);
			whichPoint.addItemListener(new ItemListener() {
				@Override
				public void itemStateChanged(ItemEvent e) {
					String cs=whichPoint.getSelectedItem();
					if(cs.startsWith("Which Point")) {
						return;
					}else if(cs.contentEquals("New Point") || cs.contentEquals("New Negative Point")) {
						if(cs.contentEquals("New Negative Point") && xys.size()==0) {
							IJ.showMessage("First point must not be negative");
							whichPoint.select(0);
							return;
						}
						pointIsNeg[xys.size()]=cs.contentEquals("New Negative Point");
						setPointIndex(xys.size());
						resetWP(xys.size()+1);
					}else if(cs.contentEquals("Delete Point")) {
						String[] items=new String[xys.size()];
						for(int i=0;i<items.length;i++)items[i]=""+(i+1);
						GenericDialog gd=new GenericDialog("Delete Point");
						gd.addChoice("Which Point to delete?",items	, items[items.length-1]);
						gd.showDialog();
						if(gd.wasCanceled()) {whichPoint.select(0); return;}
						int ind=gd.getNextChoiceIndex();
						xys.remove(ind);
						if(pointIndex>=xys.size())pointIndex=xys.size()-1;
						resetWP(xys.size());
						for(int i=0;i<curWand.length;i++) {
							if(curWand[i]!=null) {
								curWand[i].deletePoint(ind);
								if(i==(simp.getT()-1))curWand[i].showBoth();
							}
						}
					} else {
						setPointIndex(whichPoint.getSelectedIndex()-1);
						resetWP(xys.size());
					}
				}
			});
			add(whichPoint, c);
			c.gridx++; //2
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("autoacceptall");
			b.setLabel("Auto accept all");
			b.addActionListener(l);
			add(b,c);
			c.gridx++; //3
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("done");
			b.setLabel("Stop TCT");
			b.addActionListener(l);
			add(b,c);
			c.gridx=0;
			c.gridy++;
			Checkbox cb=new Checkbox("Show Each Point Roi");
			cb.setFocusable(false);
			cb.setState(showEachPointRoi);
			cb.addItemListener(new ItemListener() {
				@Override
				public void itemStateChanged(ItemEvent e) {
					showEachPointRoi=cb.getState();
					if(!showEachPointRoi)soverlay.clear();
				}
			});
			add(cb,c);
			c.gridx++;
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("addDirectRoi");
			b.setLabel("Add direct ROI");
			b.addActionListener(l);
			add(b,c);
			c.gridx++;
			b=new Button();
			b.setFocusable(false);
			b.setActionCommand("deleteDirectRoi");
			b.setLabel("Delete direct ROI");
			b.addActionListener(l);
			add(b,c);

			c.gridwidth=5; c.gridx=0; c.gridy++; //2
			ta=new TextArea("",15,90,TextArea.SCROLLBARS_NONE);
			ta.setEditable(false);
			ta.setFocusable(false);
			add(ta,c);

			pack();
		}

		/**
		 * Set (0-index) line of text area with text
		 * @param line
		 * @param text
		 */
		public void setTextLine(TctLines tctline, String text) {
			int line=tctline.ordinal();
			String[] tatext=ta.getText().split("\n");
			String out="";
			final int max=Math.max(line,tatext.length);
			for(int i=0;i<=max;i++) {
				if(i==line)out+=text+"\n";
				else if(i<tatext.length)out+=tatext[i]+"\n";
				else out+="\n";
			}
			ta.setText(out);

		}

		public void printOutput(String output) {
			if(output==null)return;
			String[] tl=output.split("\t");
			if(tl.length<12)return;
			setTextLine(TctLines.OUTPUT, "Lbl: "+tl[0]+" Cell: "+tl[2]+" Fr: "+tl[10]+" Area: "+AJ_Utils.parseDoubleTP(tl[3],2)+" Mean: "+AJ_Utils.parseDoubleTP(tl[8],2)+
					" X: "+AJ_Utils.parseDoubleTP(tl[6],2)+" Y: "+AJ_Utils.parseDoubleTP(tl[7],2)+" thresh: "+tl[11]);
		}

		public void resetWP(int max) {
			whichPoint.removeAll();
			whichPoint.add("Which Point: "+(pointIndex+1));
			for(int i=0;i<Math.max(max,xys.size());i++)whichPoint.add(""+(i+1)+(pointIsNeg[i]?"-":""));
			whichPoint.add("New Point");
			whichPoint.add("New Negative Point");
			whichPoint.add("Delete Point");
			whichPoint.select(0);
		}

	}

	private void setPointIndex(int i) {
		pointIndex=i;
		curX=-1; curY=-1;
		tctpanel.setTextLine(TctLines.CURRENTPOINT, "Point:"+(pointIndex+1)+" x:"+curX+" y:"+curY+" z:"+simp.getZ()+" fr:"+simp.getT());
	}


	private void simpleUpdateCurWand() {
		if(xys.size()==0 || xys.get(0).x==-1)return;

		int cfr=simp.getFrame();
		int i=cfr-1;
		if(curWand[i]==null)curWand[i]=new AutoWandRoi(i+1);
		else curWand[i].setThresh(pointIndex,currentThresh[pointIndex]);
		curWand[i].showBoth();
		if(sucwt!=null) {
			sucwt.stopThis();
			try {
				sucwt.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		sucwt=new MyThread() {
			public void run() {
				for(int i=0;i<simp.getNFrames();i++) {
					if(shouldStop.get()){
						if(DEBUG)IJ.log("broke simpleUpdateCurWand");
						break;
					}
					if(i==(cfr-1))continue;
					if(curWand[i]!=null) curWand[i].setThresh(pointIndex,currentThresh[pointIndex]);
				}
			}
		};
		sucwt.run();
	}

	private void updateCurWand(boolean clear) {
		final int fr=simp.getFrame();
		if(updateCurWandThread!=null) {
			updateCurWandThread.stopThis();
			try {
				updateCurWandThread.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		updateCurWandThread=new MyThread() {

			public void run() {
				for(int i=0; i<simp.getNFrames(); i++) {
					if(shouldStop.get()) {
						if(DEBUG)IJ.log("broke updateCurWand");
						break;
					}
					if(i==(fr-1))continue;
					if(curWand[i]==null) {
						curWand[i]=new AutoWandRoi(getLastAWR(i+1), i+1);
					}else if(!curWand[i].accepted || UPDATE_ACCEPTED) {// 
						if(clear)curWand[i].resetPoints();
						curWand[i].findRoiCloseTo(getLastAWR(i+1));
					}
					//IJ.log("curWand"+i+" "+curWand[i]);
				}
			}
		};
		updateCurWandThread.run();
	}

	private AutoWandRoi getLastAWR(int frame) {
		AutoWandRoi prevCw=null;
		for(int i=((frame>1)?(frame-2):0);i>=0;i--) {
			if(curWand[i]!=null && curWand[i].probable) {prevCw=curWand[i]; break;}
		}
		if(prevCw==null){
			for(int i=frame;i<simp.getNFrames();i++) {
				if(curWand[i]!=null && curWand[i].probable) {prevCw=curWand[i]; break;}
			}
		}
		return prevCw;
	}

	class MyThread extends Thread{
		AtomicBoolean shouldStop=new AtomicBoolean();
		public void stopThis() {
			shouldStop.set(true);
		}
	}

	private void autoAcceptAll() {
		for(int i=0;i<simp.getNFrames();i++) {
			AutoWandRoi cw=curWand[i];
			if(cw.probable && !cw.accepted) {
				cw.accept();
				postFirstAccept[0]=true;
			}
		}
		goToNext(1);
	}

	private void goToNext(int start) {
		for(int i=(start-1);i<simp.getNFrames();i++) {
			if(curWand[i]==null || !curWand[i].accepted) {simp.setPosition(simp.getC(), simp.getZ(), i+1);return;}
		}
		for(int i=0;i<start;i++) {
			if(curWand[i]==null || !curWand[i].accepted) {simp.setPosition(simp.getC(), simp.getZ(), i+1);return;}
		}
		int fr=simp.getT()+1;
		if(fr>simp.getNFrames())fr=1;
		simp.setPosition(simp.getC(), simp.getZ(), fr);
	}

	public void finish(){
		done=true;
	}

	private void addDirectRoi() {
		GenericDialog gd=new GenericDialog("New Direct Roi");
		gd.addChoice("New Roi should be:", DirectRoiTypes.getStrings(), DirectRoiTypes.SUBTRACTIVE.getString());
		gd.showDialog();
		if(!gd.wasCanceled()) {
			String choice=gd.getNextChoice();
			int fr=simp.getT()-1;
			Overlay temp=new Overlay();
			Roi curRoi=simp.getRoi();
			curRoi.setStrokeColor(addColor);
			temp.add(curRoi);
			simp.deleteRoi();
			simp.setOverlay(temp);
			Color def=Roi.getColor();
			Roi.setColor(DirectRoiTypes.SUBTRACTIVE.getString().equals(choice)?subColor:addColor);
			WaitForUserDialog wd=new WaitForUserDialog("Add Direct Roi", "Draw Direct Roi then hit Enter");
			wd.show();
			Roi.setColor(def);
			acceptedROI=false;
			Roi addRoi=simp.getRoi();
			simp.deleteRoi();
			simp.setOverlay(soverlay);
			if(addRoi==null) {
				IJ.log("No Roi detected, no direct Roi added");
				simp.setRoi(curRoi);
			}else {
				addRoi.setName(choice);
				directRois.add(addRoi);
				if(showEachPointRoi)addRoiToOverlay(addRoi, 0, 0, 0);
				if(curWand[fr]!=null && (!curWand[fr].accepted || UPDATE_ACCEPTED))curWand[fr].updateWandRoi();
				updateCurWand(false);
			}
			if(curWand[fr]!=null)curWand[fr].showBoth();
		}
		addDirectRoi=false;
	}

	public void deleteDirectRoi() {
		Roi curRoi=simp.getRoi();
		Roi roi=directRois.get(directRois.size()-1);
		if("sub".equals(roi.getName()))roi.setStrokeColor(subColor);
		simp.setRoi(roi);
		String[] items=new String[directRois.size()];
		for(int i=0;i<items.length;i++)items[i]=""+(i+1);
		GenericDialog gd=new GenericDialog("Delete Direct Roi");
		gd.addChoice("Which Direct Roi to delete?",items, items[items.length-1]);
		Choice choice=(Choice)gd.getChoices().get(0);
		choice.addItemListener(new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent e) {
				Roi roi=directRois.get(choice.getSelectedIndex());
				if("sub".equals(roi.getName()))roi.setStrokeColor(subColor);
				simp.setRoi(roi);
				simp.updateAndDraw();
			}

		});
		gd.showDialog();
		if(curRoi!=null)simp.setRoi(curRoi);
		if(!gd.wasCanceled()) {
			int ind=gd.getNextChoiceIndex();
			soverlay.remove(directRois.get(ind));
			directRois.remove(ind);
			int fr=simp.getT()-1;
			if(curWand[fr]!=null && (!curWand[fr].accepted || UPDATE_ACCEPTED)){
				curWand[fr].updateWandRoi();
				curWand[fr].showBoth();
			}
			updateCurWand(false);
		}
		deleteDirectRoi=false;
	}

	public synchronized void mouseWheelMoved(MouseWheelEvent e) {
		if(addDirectRoi || deleteDirectRoi)return;
		int rotation = e.getWheelRotation();
		int amount = e.getScrollAmount();
		boolean ctrl = (e.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK)!=0;

		/*
			IJ.log("mouseWheelMoved: "+e);
			IJ.log("  type: "+e.getScrollType());
			IJ.log("  ctrl: "+ctrl);
			IJ.log("  rotation: "+rotation);
			IJ.log("  amount: "+amount);
		 */
		if (!ctrl) {
			//if (amount<1) amount=1;
			amount=1;
			if (rotation==0)
				return;
			ImageProcessor sip=simp.getProcessor();
			if(currentThresh[pointIndex]<20)wheelfactor=1;
			currentThresh[pointIndex]+=(rotation*amount*wheelfactor);
			sip.setThreshold(currentThresh[pointIndex], (double) 65535, MYLUT);
			tctpanel.setTextLine(TctLines.THRESH, "Thresh: "+currentThresh[pointIndex]);
			simp.updateAndDraw();
			simpleUpdateCurWand();
		} else{
			simp.getWindow().mouseWheelMoved(e);
		}
	}

	public void actionPerformed(ActionEvent event){ 
		if(DEBUG)IJ.log("Button pressed"); 
	} 

	public void mousePressed(MouseEvent e) { 
		if(IJ.altKeyDown() || IJ.shiftKeyDown() || IJ.spaceBarDown() || addDirectRoi || deleteDirectRoi) return;
		int flags = e.getModifiersEx();
		if(DEBUG) IJ.log("flags: "+flags+" Mask1: "+InputEvent.BUTTON1_DOWN_MASK+" Mask2: "+InputEvent.BUTTON2_DOWN_MASK+" Mask3: "+InputEvent.BUTTON3_DOWN_MASK);
		if(DEBUG) IJ.log("buttonPressed: "+(((flags&InputEvent.BUTTON1_DOWN_MASK)!=0)?"1":"")+(((flags&InputEvent.BUTTON2_DOWN_MASK)!=0)?"2":"")+(((flags&InputEvent.BUTTON3_DOWN_MASK)!=0)?"3":""));
		if((flags & InputEvent.BUTTON1_DOWN_MASK) !=0){
			Point xy = simp.getCanvas().getCursorLoc(); //because x,y position isn't exactly x,y in image
			curX=xy.x; curY=xy.y;
			if(DEBUG) IJ.log("Position: "+xy.x+" "+xy.y);
			buttonpress[0]=true;
		}
		if((flags & InputEvent.BUTTON2_DOWN_MASK) !=0)buttonpress[1]=true;
		if((flags & InputEvent.BUTTON3_DOWN_MASK) !=0) {buttonpress[2]=true; acceptedROI=true;}
	}
	public void mouseReleased(MouseEvent e) {
		if(addDirectRoi || deleteDirectRoi) return;
		if(IJ.altKeyDown() || IJ.shiftKeyDown()) {
			int fr=simp.getT();
			if(curWand[fr-1]!=null)curWand[fr-1].setRoi(simp.getRoi());
			return;
		}
		if(DEBUG) IJ.log("buttonReleased: "+e.getModifiersEx()+" button: "+e.getButton());
		if(e.getButton()==MouseEvent.BUTTON1)buttonpress[0]=false;
		if(e.getButton()==MouseEvent.BUTTON2) {
			IJ.run("Previous Slice [<]");
			buttonpress[1]=false;
		}
		if(e.getButton()==MouseEvent.BUTTON3) {
			buttonpress[2]=false;
			if(goback==1){IJ.run("Previous Slice [<]");}
			//else{acceptedROI=true;}
		}
	} 
	public void mouseExited(MouseEvent e) {} 
	public void mouseClicked(MouseEvent e) {}	
	public void mouseEntered(MouseEvent e) {} 


	public void keyPressed(KeyEvent e) {
		int keyCode = e.getKeyCode();
		//char keyChar = e.getKeyChar();
		//int flags = e.getModifiers();
		//IJ.log("keyPressed: keyCode=" + keyCode + " (" + KeyEvent.getKeyText(keyCode) + ")");
		if(keyCode==32) { //space key
			spacepress=true;
		}
	}

	public void keyReleased(KeyEvent e) {
		int keyCode = e.getKeyCode();
		//char keyChar = e.getKeyChar();
		if(keyCode==32) { //space key
			if(!buttonpress[0]){ //mouse button while spacebar pressed
				if(goback==2){IJ.run("Previous Slice [<]");}
				else{acceptedROI=true;}
			}
			spacepress=false;
		}
		if(keyCode==KeyEvent.VK_F1){
			changeLutType();
		}
		if(keyCode==KeyEvent.VK_F2){
			changeGoBack();
		}
		if(keyCode==KeyEvent.VK_F3){
			askCellLabel();
		}
		if(keyCode==KeyEvent.VK_F4){gocellcomplete=true;}
		if(keyCode==KeyEvent.VK_UP || keyCode==KeyEvent.VK_DOWN){
			changeWheelFactor((keyCode==KeyEvent.VK_UP)?1:-1);
		}
		if(keyCode==27) {done=true;}
	}

	public void keyTyped(KeyEvent e) {}

	private void changeLutType() {
		MYLUT++;
		if(MYLUT>3)MYLUT=0;
		if(MYLUT==2) {
			ImageProcessor sip=simp.getProcessor();
			sip.setMinAndMax(sip.getMin(),sip.getMax());
		}
	}

	private void changeGoBack() {
		goback++; if(goback>2)goback=0;
		if(goback==1) tctpanel.setTextLine(TctLines.INFO, "Rightclick now goes back a frame");
		else if(goback==2)tctpanel.setTextLine(TctLines.INFO, "Space key now goes back a frame");
		else tctpanel.setTextLine(TctLines.INFO, "No back button");
	}

	/**
	 * upOrDown must be 1 or -1
	 * @param upOrDown
	 */
	private void changeWheelFactor(int upOrDown) {
		int i=0;
		for(;i<whfacs.length; i++) {
			if(wheelfactor<=whfacs[i]) break;
		}
		int newi=i+upOrDown;
		if(newi>whfacs.length)newi=0;
		if(newi<0)newi=whfacs.length-1;
		wheelfactor=whfacs[newi];
	}

	private static final int FUZZD=3;

	public static boolean areRoisTouching(Roi one, Roi two){
		if(one==null || two==null)return false;
		Rectangle ob=one.getBounds();
		Rectangle tb=two.getBounds();

		//First, quickly rule out rois that are not close based on bounds
		boolean left,above;
		left=(ob.x+ob.width)<(tb.x+tb.width);
		above=(ob.y+ob.height)<(tb.y+tb.height);
		//IJ.log("OB: "+ob.x+","+ob.y+","+ob.width+","+ob.height+" TB: "+tb.x+","+tb.y+","+tb.width+","+tb.height+" "+(left?"Left,":"Right,")+(above?"Above":"Below"));
		if(left?((ob.x+ob.width+FUZZD)<(tb.x)):((tb.x+tb.width+FUZZD)<ob.x))return false;
		if(above?((ob.y+ob.height+FUZZD)<(tb.y)):((tb.y+tb.height+FUZZD)<ob.y))return false;
		//IJ.log("Crossed bounds");

		/* In case rois are PolygonRoi instead of ShapeRoi
		Polygon onep=one.getPolygon();
		Polygon twop=two.getPolygon();
		for(int i=0;i<onep.npoints;i++) {
			int x=onep.xpoints[i]+(left?FUZZD:-FUZZD), y=onep.ypoints[i]+(above?FUZZD:-FUZZD);
			for(int j=0;j<twop.npoints;j++) {
				int xc=twop.xpoints[j],yc=twop.ypoints[j];
				//IJ.log(((i==0&&j==0)?"":"\\Update:")+"One: "+x+","+y+"  Two: "+xc+","+yc);
				if((left?(x>=xc):(xc>=x)) && (above?(y>=yc):(yc>=y)))return true;
			}
		}
		 */

		//Then check if their areas are really within FUZZD
		ShapeRoi sroi=(enlargeRoi(one,FUZZD)).and(enlargeRoi(two,FUZZD));
		return (sroi!=null && (sroi.getPolygon().npoints>0));
	}

	public static ShapeRoi enlargeRoi(Roi roi, int enlarge) {
		return (new ShapeRoi(RoiEnlarger.enlarge(roi,enlarge)));
	}

	/**
	 * Splits a non-contiguous ShapeRoi and returns the biggest one
	 * according to the bounding rectangle (not actual area).
	 * @param roi
	 * @return
	 */
	public static Roi biggestNGRoi(Roi roi) {
		Roi result=roi;
		if((roi instanceof ShapeRoi)) {
			Roi[] rois=((ShapeRoi)roi).getRois();
			if(rois.length>1) {
				int maxarea=0, index=-1;
				for(int i=0;i<rois.length;i++) {
					Rectangle b=rois[i].getBounds();
					int area=b.width*b.height;
					if(area>maxarea) {maxarea=area; index=i;}
				}
				result=rois[index];
			}
		}
		return result;
	}


	private void addRoiToOverlay(Roi roi, int ch, int sl, int fr) {
		if(roi!=null) {
			roi.setStrokeColor(DirectRoiTypes.SUBTRACTIVE.getString().equals(roi.getName())?subColor:addColor);
			roi.setPosition(ch,sl,fr);
			soverlay.add(roi);
		}
	}

	class AutoWandRoi{
		private ImageProcessor ip;
		private int ch,sl,fr;
		private Calibration cal=simp.getCalibration();
		private Roi[] rois=new Roi[MAX_POINTS];
		private ArrayList<Double> threshs=new ArrayList<Double>();
		private ArrayList<Point> cxys=new ArrayList<Point>();
		private boolean[] cPointIsNeg=new boolean[MAX_POINTS];

		public Roi roi;
		public Roi troi=null;
		public double area;
		public double mean;
		public Point centroid;
		public String output;
		public String pointFail="";

		public boolean probable=false;
		public boolean accepted=false;

		public AutoWandRoi(Roi roi, int frame, double thresh) {
			this.roi=roi;
			this.fr=frame;
			this.ch=simp.getC();
			this.sl=simp.getZ();
			this.ip=simp.getImageStack().getProcessor(simp.getStackIndex(ch, sl, frame));
			cal=simp.getCalibration();
			cxys.add(new Point((int)roi.getXBase(),(int)roi.getYBase()));
			threshs.add(thresh);
			fillStats();
			probable=true;
		}

		public AutoWandRoi(int frame) {
			this.ip=simp.getImageStack().getProcessor(simp.getStackIndex(simp.getC(), simp.getZ(), frame));
			this.fr=frame;
			resetPoints();
			probable=true;
		}

		public AutoWandRoi(AutoWandRoi prevAWI, int frame) {
			this.fr=frame;
			this.ip=simp.getImageStack().getProcessor(simp.getStackIndex(simp.getC(), simp.getZ(), frame));
			findRoiCloseTo(prevAWI);
		}

		private Roi doWandRoi(Point p, double thresh){
			int xw=p.x, yw=p.y;
			if(xw==-1 || yw==-1) return null;

			Wand w=new Wand(ip);
			w.autoOutline(xw,yw,thresh,(double)65535);
			Roi sel=new ShapeRoi(new PolygonRoi(w.xpoints, w.ypoints, w.npoints, Roi.POLYGON));

			Rectangle sbs=sel.getBounds();
			if(sel!=null && sbs.getWidth()<(simp.getWidth()/2) && sbs.getHeight()<(simp.getHeight()/2) && sbs.getX()>0 && sbs.getY()>0 && sbs.getWidth()>5 && sbs.getHeight()>5){
				sel=ij.plugin.RoiEnlarger.enlarge(ij.plugin.RoiEnlarger.enlarge(sel,3),-3);
			}
			return sel;
		}

		private Roi doWandRoi(int index) {
			return doWandRoi(cxys.get(index), threshs.get(index));
		}

		private void fillStats() {
			if(roi!=null) { 
				ip.setRoi(roi);
				ImageStatistics imgstat=ImageStatistics.getStatistics(ip, 127, cal);
				area=imgstat.area;
				mean=imgstat.mean;
				centroid=new Point((int)(imgstat.xCentroid/cal.pixelWidth),(int)(imgstat.yCentroid/cal.pixelHeight));
			}else {

			}
		}

		/*
    	private ImageStatistics getStats(Roi roi) {
    		ip.setRoi(roi);
    		return ImageStatistics.getStatistics(ip, 127, cal);
    	}
		 */

		private void updateWandRoi() {updateWandRoi(null);}

		private void updateWandRoi(AutoWandRoi prev) {
			if(ch!=simp.getC() || sl!=simp.getZ()) {
				ch=simp.getC(); sl=simp.getZ();
				ip=simp.getImageStack().getProcessor(simp.getStackIndex(ch, sl, fr));
			}
			if(cxys.size()==0) {
				if(xys.size()==0)return;
				resetPoints();
				return;
			}
			if(showEachPointRoi) {
				for(int i=0;i<rois.length;i++) {soverlay.remove(roi);
				if(rois[i]!=null) {
					//IJ.log("remove ov roi"+i+" T"+fr); 
					soverlay.remove(rois[i]);
				}
				rois[i]=null;
				}
			}
			pointFail="";
			//boolean anyneg=false;
			for(int i=0;i<cxys.size();i++) {
				rois[i]=doWandRoi(i);
				if(i==0) {
					if(cxys.size()>1 && prev!=null && !areRoisTouching(roi, prev.roi)) {rois[0]=null; pointFail+=" "+(i+1);}
				}else {
					if(!cPointIsNeg[i]) {
						if(prev==null || areRoisTouching(rois[i],prev.roi)) {}//addRoi(rois[i]);
						else {pointFail+=" "+(i+1); rois[i]=null;}
					}else {
						//anyneg=true;
						if(prev==null || (prev.rois.length>i && areRoisTouching(rois[i],prev.rois[i]))) {}//subtractRoi(rois[i]);
						else {pointFail+=" "+(i+1); rois[i]=null;}
					}
				}
			}
			buildRoi();
			/*
			 * can't just keep biggest roi because sometimes added points are not contiguous
			if(anyneg) {
				roi=biggestNGRoi(roi);
			}
			 */
			fillStats();
		}

		private void buildRoi() {
			roi=null;
			for(int i=0;i<rois.length;i++) {
				if(rois[i]!=null) {
					if(cPointIsNeg[i]) {rois[i].setName(DirectRoiTypes.SUBTRACTIVE.getString()); subtractRoi(rois[i]);}
					else addRoi(rois[i]);
					if(showEachPointRoi)addRoiToOverlay(rois[i], ch, sl, fr);
				}
			}
			for(int i=0;i<directRois.size();i++) {
				Roi temp=directRois.get(i);
				if(temp!=null) {
					boolean isNeg=DirectRoiTypes.SUBTRACTIVE.getString().equals(temp.getName());
					if(isNeg)subtractRoi(temp);
					else if(DirectRoiTypes.ONLY_WITHIN.getString().equals(temp.getName())) {
						andRoi(temp);
					}else addRoi(temp);
					//if(showEachPointRoi)addRoiToOverlay(temp, isNeg, ch, sl, fr);
				}
			}
		}

		//private void setRoi(Roi roi, int index) {
		//	rois[index]=roi;
		//	buildRoi();
		//	fillStats();
		//}

		private void setRoi(Roi roi) {
			if(roi==this.roi)return;
			this.roi=roi;
			fillStats();
		}

		public void addPoint(Point point, double thresh, boolean isNeg) {
			cxys.add((Point)point.clone());
			threshs.add(thresh);
			cPointIsNeg[threshs.size()-1]=isNeg;
		}

		public void setPoint(int index, Point point, double thresh) {
			if(index>=cxys.size()) {IJ.showMessage("setPoint index out of range");return;}
			//if(index==cxys.size())addPoint(point, thresh);
			else {
				cxys.set(index, (Point)point.clone());
				threshs.set(index, thresh);
			}
		}

		public void updatePoint(int index, Point point, double thresh) {
			if(index>=cxys.size()) {IJ.showMessage("updatePoint index out of range");return;}
			setPoint(index,point,thresh);
			updateWandRoi();
		}

		public void updatePoint() {
			if(pointIndex>=cxys.size()) {
				for(int i=cxys.size();i<=pointIndex;i++) {addPoint(xys.get(i), currentThresh[i], pointIsNeg[i]);}
			}else if(pointIndex<cxys.size())setPoint(pointIndex, xys.get(pointIndex), currentThresh[pointIndex]);
			updateWandRoi();
		}

		public void deletePoint(int index) {
			if(index<cxys.size() && index < threshs.size()) {
				cxys.remove(index); threshs.remove(index);
			}
			updateWandRoi();
		}

		public void resetPoints() {
			for(int i=0;i<xys.size();i++) {
				if(i==cxys.size()) addPoint(xys.get(i), currentThresh[i], pointIsNeg[i]);
				else setPoint(i, xys.get(i), currentThresh[i]);
			}
			updateWandRoi();
		}

		private void addRoi(Roi addroi) {
			if(roi==null) {roi=addroi; return;}
			if(addroi==null)return;
			if(! (addroi instanceof ShapeRoi))addroi=new ShapeRoi(addroi);
			if(! (roi instanceof ShapeRoi))roi=new ShapeRoi(roi);
			roi=((ShapeRoi)roi).or((ShapeRoi)addroi);
		}
		
		private void andRoi(Roi androi) {
			if(roi==null) {return;}
			if(androi==null)return;
			if(! (androi instanceof ShapeRoi))androi=new ShapeRoi(androi);
			if(! (roi instanceof ShapeRoi))roi=new ShapeRoi(roi);
			roi=((ShapeRoi)roi).and((ShapeRoi)androi);
		}

		private void subtractRoi(Roi subroi) {
			if(roi==null)return;
			if(subroi==null)return;
			if(! (subroi instanceof ShapeRoi))subroi=new ShapeRoi(subroi);
			if(! (roi instanceof ShapeRoi))roi=new ShapeRoi(roi);
			roi=((ShapeRoi)roi).not((ShapeRoi)subroi);
		}

		public Roi getRoi() {return roi;}
		public Roi getTRoi() {return troi;}

		public double getThresh(int index) {return threshs.get(index);}

		private void findRoiCloseTo(AutoWandRoi prev) {
			probable=false;
			if(prev==null)return;
			if(prev.area==0)return;
			if(cxys.size()==0) {
				resetPoints();
				//updateWandRoi(prev);
			}else if(cxys.size()<xys.size()) {
				for(int i=cxys.size();i<xys.size();i++)addPoint(xys.get(i),currentThresh[i], pointIsNeg[i]);
				//updateWandRoi(prev);
			}
			updateWandRoi(prev);
			Point c=new Point(xys.get(0).x,xys.get(0).y);
			ArrayList<Double> ranks=new ArrayList<Double>();
			ArrayList<Point> pts=new ArrayList<Point>();
			for(int i=0;i<12;i++) {
				if(roi!=null) {
					double distDiff=Math.hypot(((double)centroid.x*cal.pixelWidth-(double)prev.centroid.x*cal.pixelWidth),((double)centroid.y*cal.pixelHeight-(double)prev.centroid.y*cal.pixelHeight));
					double radDiff=Math.abs((Math.sqrt(area/Math.PI))-(Math.sqrt(prev.area/Math.PI)));
					double rad=(Math.sqrt(prev.area/Math.PI));
					if(distDiff<rad && radDiff<(rad/3)){
						probable=true; 
						if(DEBUG)IJ.log("frct"+fr+": acc d"+IJ.d2s(distDiff,3)+" r"+IJ.d2s(radDiff,3)+" rad"+IJ.d2s(rad,3)+" rank"+(distDiff*3/rad+radDiff/rad));
						return;
					}
					ranks.add(distDiff*3/rad+radDiff/rad);
					pts.add((Point)c.clone());
					if(DEBUG) {
						//if(roi!=null)IJ.log("frct:"+fr+" rej "+(Math.hypot((centroid.x-prev.centroid.x),(centroid.y-prev.centroid.y))>(Math.sqrt(prev.area/Math.PI)/cal.pixelWidth)?"too far  ":"")+
						//		((area/prev.area > (1+AREATOLERANCE))?"too big   ":"")+((area/prev.area<AREATOLERANCE)?"too small  ":""));
						//else IJ.log("frct:"+fr+" rej null roi");
					}
				}else {

				}
				if(i==6) {c.x=prev.centroid.x; c.y=prev.centroid.y;} 
				else {c.x-=10; c.y-=10;}
				setPoint(0, c, currentThresh[0]);
				updateWandRoi(prev);
			}
			int index=0;
			double minRank=65535.0;
			for(int i=0; i<ranks.size();i++) { if(ranks.get(i)<minRank) {minRank=ranks.get(i); index=i;}}
			if(index<pts.size())
				setPoint(0,pts.get(index), currentThresh[0]);
			else 
				setPoint(0,c, currentThresh[0]);
			updateWandRoi();
			//if(DEBUG)IJ.log("findRoiCloseTo: "+attempts);
		}

		public void setThresh(int index, double thresh) {
			if(index>=cxys.size()) {
				if(DEBUG)IJ.log("TCT ERROR: Can't set thresh for index: "+index);
				return;
			}
			setPoint(index, cxys.get(index), thresh);
			updateWandRoi();
		}

		public void showBoth() {
			if(simp.getRoi()!=roi){
				simp.setRoi(roi);
				timp.setRoi((Roi)roi.clone());
			}
		}

		public void accept() {
			if(roi==null) {IJ.log("TCT Error: No Selection"); return;}
			probable=true;
			Overlay tov=timp.getOverlay();
			if(tov==null) {tov=new Overlay(); tov.setFillColor(new Color(255,100,0)); timp.setOverlay(tov);}
			if(troi!=null) {
				tov.remove(troi);
				timp.deleteRoi();
			}
			troi=(Roi)roi.clone();
			troi.setImage(timp);
			troi.setPosition(1, labelsl, fr);
			troi.setFillColor(new Color(255,100,0));
			timp.setPosition(1,labelsl,fr);
			tov.add(troi);
			timp.updateAndRepaintWindow();

			double perimeter,circularity;
			perimeter=roi.getLength();
			circularity = perimeter==0.0?0.0:4.0*Math.PI*(area/(perimeter*perimeter));
			output=""+cellLabel+"\t"+results.roin+"\t"+celln+"\t"+area+"\t"+perimeter+"\t"+circularity+"\t"+centroid.x*cal.pixelWidth+"\t"+centroid.y*cal.pixelHeight+"\t"+mean+"\t"+sl+"\t"+fr+"\t"+getThresh(0)+"\t"+times[fr-1];
			accepted=true;
		}

		public void draw() {
			ImageStack ist=timp.getImageStack();
			ImageProcessor tip1=ist.getProcessor(timp.getStackIndex(1, labelsl, fr));
			ImageProcessor tip2=ist.getProcessor(timp.getStackIndex(2, labelsl, fr));
			if(troi!=null) {
				tip1.setRoi(troi);
				ImageStatistics imgstat=ImageStatistics.getStatistics(tip1, 127, timp.getCalibration());
				tip1.setColor(255);
				tip1.fill(troi);
				tip1.setColor(0);
				tip1.draw(ij.plugin.RoiEnlarger.enlarge(troi,1));
				tip1.setColor(255);

				String cl=""+celln;
				if(drawCellLabel)cl=cl.concat(" "+cellLabel);
				if(DEBUG)IJ.log("Drawing: "+fr+" "+(imgstat.xCentroid)+" "+(imgstat.yCentroid)+"    "+cl);
				tip2.setColor(255);
				tip2.drawString(cl, (int)imgstat.xCentroid-5, (int)imgstat.yCentroid-10);
				results.append(output);
			}
		}

	}

}
